# /// script
# dependencies = [
#   "julax",
#   "opencv-python",
#   "tensorflow-datasets>=4.9.9",
# ]
#
# [tool.uv.sources]
# julax = { path = "../", editable = true }
# ///

import logging


import grain
import jax
from jax.nn.initializers import truncated_normal
import optax
import tensorflow_datasets as tfds

from julax import (
    Chain,
    DoEveryNSteps,
    Experiment,
    Learner,
    Linear,
    Param,
    State,
    Trainer,
    default_observer,
    test_mode,
)

from absl import logging as absl_logging

logging.root.setLevel(logging.INFO)
absl_logging.use_python_logging()


def evaluate(x: Experiment, p: Param, s: State):
    dataset = (
        grain.MapDataset.source(tfds.data_source("mnist", split="test"))
        .batch(32, drop_remainder=True)
        .map(
            lambda x: {
                "feature": x["image"].reshape(32, -1),
                "label": x["label"],
            }
        )
        .to_iter_dataset()
    )
    model = x.trainer.learner.model
    param = p["trainer"]["learner"]["model"]
    state = test_mode(s["trainer"]["learner"]["model"])
    n_correct, n_total = 0, 0
    for batch in iter(dataset):
        ŷ, _ = model(batch["feature"], param, state)
        n_correct += (ŷ.argmax(axis=1) == batch["label"]).sum().item()
        n_total += 32
    acc = n_correct / n_total

    logging.info(f"Accuracy at step {s['step']}: {acc}")


E = Experiment(
    name="mnist",
    trainer=Trainer(
        learner=Learner(
            model=Chain(
                Linear(
                    in_dim=784,
                    out_dim=512,
                    w_init=truncated_normal(),
                ),
                jax.nn.relu,
                Linear(
                    in_dim=512,
                    out_dim=512,
                    w_init=truncated_normal(),
                ),
                jax.nn.relu,
                Linear(
                    in_dim=512,
                    out_dim=10,
                    w_init=truncated_normal(),
                ),
            ),
            loss_fn=optax.softmax_cross_entropy_with_integer_labels,
        ),
        optimizer=optax.sgd(0.01),
    ),
    dataset=grain.MapDataset.source(tfds.data_source("mnist", split="train"))
    .seed(seed=45)
    .shuffle()
    .batch(32, drop_remainder=True)
    .map(
        lambda x: {
            "feature": x["image"].reshape(32, -1),
            "label": x["label"],
        }
    )
    .slice(slice(1000))
    .to_iter_dataset(),
    observer=default_observer() * DoEveryNSteps(evaluate, n=100),
)

E.run()
E.close()



from functools import cached_property
import jax
from jax.sharding import PartitionSpec, Mesh

from julax.utils import create_mesh

from .core import PRNG, LayerBase, Trainer, State, Param, PyTree
import grain

import orbax.checkpoint as ocp

import logging

from pydantic import Field, computed_field

from .observers import default_observer, ObserverBase

logger = logging.getLogger(__name__)


class Experiment(LayerBase):
    name: str = "mnist"

    seed: int = 0
    trainer: Trainer

    dataset: grain.IterDataset

    max_steps: int | None = None
    batch_axis_names: list[str] = ["data"]
    mesh_shape: dict[str, int] = {"data": -1}

    checkpoint_manager: ocp.CheckpointManager | None = None
    observer: ObserverBase = Field(default_factory=default_observer)

    @computed_field
    @cached_property
    def mesh(self) -> Mesh:
        return create_mesh(self.mesh_shape)

    def state(self, rng: PRNG) -> State:
        return State(input=iter(self.dataset), step=0)

    def forward(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]:
        P, S = self.trainer(x, p["trainer"], s["trainer"])
        return Param(trainer=P), State(trainer=S, input=s["input"], step=s["step"] + 1)

    def save(self, p: Param, s: State):
        if self.checkpoint_manager:
            self.checkpoint_manager.save(
                s["step"],
                args=ocp.args.Composite(
                    param=ocp.args.PyTreeSave(item=p),
                    state_trainer=ocp.args.PyTreeSave(item=s["trainer"]),
                    state_dataset_iter=grain.checkpoint.CheckpointSave(item=s["input"]),
                ),
            )

    def restore(self) -> tuple[Param, State]:
        p, s = self.init(self.seed)
        if self.checkpoint_manager is None:
            return p, s
        try:
            restored = self.checkpoint_manager.restore(
                step=None,
                args=ocp.args.Composite(
                    param=ocp.args.PyTreeRestore(
                        item=p,
                        restore_args=ocp.checkpoint_utils.construct_restore_args(p),
                    ),
                    state_trainer=ocp.args.PyTreeRestore(
                        item=s["trainer"],
                        restore_args=ocp.checkpoint_utils.construct_restore_args(
                            s["trainer"]
                        ),
                    ),
                    state_dataset_iter=grain.checkpoint.CheckpointRestore(
                        item=s["input"]
                    ),
                ),
            )
            param = restored["param"]
            state_trainer = restored["state_trainer"]
            state_dataset_iter = restored["state_dataset_iter"]
            return param, State(input=state_dataset_iter, trainer=state_trainer)
        except FileNotFoundError:
            logger.warning(
                f"No checkpoints found under {self.checkpoint_manager.directory} ! Experiment initialized with seed {self.seed}"
            )
            return p, s

    def close(self):
        if self.checkpoint_manager:
            self.checkpoint_manager.close()

    def run(self) -> tuple[Param, State]:
        with self.mesh as mesh:
            p, s = self.restore()
            self.observer(self, p, s)

            for x_local in s["input"]:
                if self.max_steps is not None and s["step"] >= self.max_steps:
                    break
                x = jax.make_array_from_process_local_data(
                    sharding=jax.sharding.NamedSharding(
                        mesh, PartitionSpec(self.batch_axis_names)
                    ),
                    local_data=x_local,
                )
                p, s = self(x, p, s)

                self.observer(self, p, s)
                self.save(p, s)

            return p, s



import logging
import time
from typing import Protocol

import jax
from .core import Param, State

logger = logging.getLogger(__name__)


class Observer(Protocol):
    def __call__(self, x, p: Param, s: State): ...


class ObserverBase:
    def __call__(self, x, p: Param, s: State):
        raise NotImplementedError

    def __mul__(self, other: Observer) -> "CompositeObserver":
        return CompositeObserver([self, other])

    def __rmul__(self, other: Observer) -> "CompositeObserver":
        return CompositeObserver([other, self])


class DoEveryNSteps(ObserverBase):
    def __init__(self, observer: Observer, n: int = 1):
        self.n = n
        self.observer = observer

    def __call__(self, x, p: Param, s: State):
        step = s["step"]
        if step % self.n == 0:
            return self.observer(x, p, s)


class DoAtStep0(ObserverBase):
    def __init__(self, observer: Observer):
        self.observer = observer

    def __call__(self, x, p: Param, s: State):
        step = s["step"]
        if step == 0:
            return self.observer(x, p, s)


class CompositeObserver(ObserverBase):
    def __init__(self, observers: list[Observer]):
        self.observers = []
        for obs in observers:
            if isinstance(obs, CompositeObserver):
                self.observers.extend(obs.observers)
            else:
                self.observers.append(obs)

    def __call__(self, x, p: Param, s: State):
        for observer in self.observers:
            observer(x, p, s)


class LossLogger(ObserverBase):
    def __call__(self, x, p: Param, s: State):
        loss = s["trainer"]["loss"]
        step = s["step"]
        jax.debug.print("Step {step}: loss={loss}", step=step, loss=loss)


class StepTimeLogger(ObserverBase):
    def __init__(self, n: int = 100):
        self.n = n
        self.last_time = None
        self.step_count = 0

    def __call__(self, x, p: Param, s: State):
        if self.last_time is None:
            self.last_time = time.perf_counter()
            self.step_count = 0
            return

        self.step_count += 1

        if self.step_count % self.n == 0:
            now = time.perf_counter()
            avg_time = (now - self.last_time) / self.step_count
            step = s["step"]
            logger.info(
                f"Step {step}: avg step time over last {self.step_count} steps: {avg_time:.6f}s"
            )
            self.last_time = now
            self.step_count = 0


def default_observer() -> CompositeObserver:
    return DoEveryNSteps(LossLogger(), n=10) * StepTimeLogger()



from typing import TypeAlias, Any
from pydantic import ConfigDict, RootModel
from jax import Array
from jax.sharding import PartitionSpec
import plum

PRNG: TypeAlias = Array
PyTree: TypeAlias = Any
OutShardingType: TypeAlias = PartitionSpec | None

# TODO: isinstance(jnp.dtype, jnp.float32) fails
Dtype: TypeAlias = Any

dispatch = plum.Dispatcher(warn_redefinition=True)


class FrozenDict(RootModel[dict]):
    model_config = ConfigDict(frozen=True)

    def __getitem__(self, item):
        return self.root[item]

    def __iter__(self):
        return iter(self.root)

    def keys(self):
        return self.root.keys()

    def values(self):
        return self.root.values()

    def items(self):
        return self.root.items()

    def __len__(self):
        return len(self.root)

    def __hash__(self):
        return hash(frozenset(self.root.items()))

    def __eq__(self, other):
        if isinstance(other, FrozenDict):
            return self.root == other.root
        return self.root == other



from abc import ABC, abstractmethod
import plum
from functools import partial
from typing import Annotated, Callable, TypeAlias, Any

import optax
from pydantic import BaseModel, BeforeValidator, ConfigDict, ValidationError


#####

import jax
import jax.numpy as jnp
from jax import jit, value_and_grad

#####

from julax.base import PRNG, PyTree, dispatch

# TODO: use RootModel[dict] for better customization
# Or maybe SimpleNamespace?
Param: TypeAlias = dict
State: TypeAlias = dict

#####


class LayerBase(BaseModel, ABC):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        frozen=True,
        ignored_types=(
            jax.stages.Wrapped,
            plum.function.Function,
            optax.GradientTransformation,
        ),
    )

    def sublayers(self) -> dict:
        attrs_flatten, treedef = jax.tree.flatten(
            dict(self), is_leaf=lambda x: isinstance(x, LayerBase)
        )
        masked_sublayers = jax.tree.unflatten(
            treedef, [x if isinstance(x, LayerBase) else None for x in attrs_flatten]
        )

        res = {}
        for k, v in masked_sublayers.items():
            if jax.tree.reduce(
                lambda x, y: x or y,
                v,
                None,
                is_leaf=lambda x: isinstance(x, LayerBase),
            ):
                res[k] = v
        return res

    def __getitem__(self, key: str) -> "LayerBase":
        return self.sublayers()[key]

    def _ipython_display_(self):
        from julax.pprint import pprint

        pprint(self)

    def param(self, rng: PRNG) -> Param:
        return Param()

    def param_length(self) -> int:
        return 0

    def state(self, rng: PRNG) -> State:
        return State()

    def state_length(self) -> int:
        return 0

    def numel(self) -> tuple[int, int]:
        num_params = self.param_length()
        num_states = self.state_length()

        for sublayer in self.sublayers().values():
            p, s = sublayer.numel()
            num_params += p
            num_states += s

        return num_params, num_states

    @dispatch
    def init(self, seed: int = 0) -> tuple[Param, State]:
        return self.init(jax.random.key(seed))

    @dispatch
    def init(self, rng: PRNG) -> tuple[Param, State]:
        sublayers, treedef = jax.tree.flatten(
            self.sublayers(), is_leaf=lambda x: isinstance(x, LayerBase)
        )

        sublayer_params_flatten, sublayer_stats_flatten = [], []

        for layer in sublayers:
            if layer is None:
                sublayer_params_flatten.append(None)
                sublayer_stats_flatten.append(None)
            else:
                rng, _rng = jax.random.split(rng)
                p, s = layer.init(_rng)
                sublayer_params_flatten.append(p)
                sublayer_stats_flatten.append(s)

        sublayer_params = Param(**jax.tree.unflatten(treedef, sublayer_params_flatten))
        sublayer_states = State(**jax.tree.unflatten(treedef, sublayer_stats_flatten))

        rng_p, rng_s = jax.random.split(rng)
        layer_params = self.param(rng_p)
        layer_states = self.state(rng_s)
        return self.init(layer_params, layer_states, sublayer_params, sublayer_states)

    @dispatch
    def init(
        self, layer_params, layer_states, sublayer_params, sublayer_states
    ) -> tuple[Param, State]:
        assert len(layer_params.keys() & sublayer_params.keys()) == 0
        assert len(layer_states.keys() & sublayer_states.keys()) == 0

        return (
            sublayer_params | layer_params,
            sublayer_states | layer_states,
        )

    @abstractmethod
    def forward(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]: ...

    @dispatch
    def __call__(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]:
        # TODO: make sure treedef of the returned state is not changed compared to `s`?
        return self.forward(x, p, s)

    @dispatch
    def __call__(self, x: PyTree) -> tuple[PyTree, State]:
        return self.__call__(x, *self.init())


@dispatch(precedence=1)
def to_layer(x: LayerBase):
    return x


@dispatch
def to_layer(x):
    raise ValidationError(f"Failed to convert to LayerBase: {x}")


LayerLike = Annotated[LayerBase, BeforeValidator(to_layer)]


class Learner(LayerBase):
    loss_fn: Callable[[PyTree, PyTree], Any]
    model: LayerBase
    agg: Callable = jnp.mean
    feature_name: str = "feature"
    label_name: str = "label"

    def forward(self, input: dict, p: Param, s: State) -> tuple[PyTree, State]:
        x = input[self.feature_name]
        y = input[self.label_name]
        ŷ, s["model"] = self.model(x, p["model"], s["model"])
        losses = self.loss_fn(ŷ, y)
        loss = self.agg(losses)
        return loss, s


class Trainer(LayerBase):
    learner: Learner
    optimizer: Any

    def state(self, rng: PRNG) -> State:
        return State(optimizer=None, loss=0.0)

    @dispatch
    def init(
        self, layer_params, layer_states, sublayer_params, sublayer_states
    ) -> tuple[Param, State]:
        layer_states["optimizer"] = self.optimizer.init(sublayer_params["learner"])
        return (
            sublayer_params | layer_params,
            sublayer_states | layer_states,
        )

    def forward(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]:
        loss, state = self.learner(x, p["learner"], s["learner"])
        return loss, State(learner=state, optimizer=s["optimizer"], loss=loss)

    @partial(jit, static_argnums=0, donate_argnames=("p", "s"))
    def forward_and_backward(
        self, x: PyTree, p: Param, s: State
    ) -> tuple[Param, State]:
        (_, S), grads = value_and_grad(self.forward, argnums=1, has_aux=True)(x, p, s)
        updates, S["optimizer"] = self.optimizer.update(
            grads["learner"], S["optimizer"]
        )
        P = optax.apply_updates(p["learner"], updates)
        return {"learner": P}, S

    @dispatch
    def __call__(self, x: PyTree, p: Param, s: State) -> tuple[Param, State]:
        return self.forward_and_backward(x, p, s)



from .core import (
    LayerBase,
    LayerLike,
    Learner,
    PRNG,
    Param,
    PyTree,
    State,
    Trainer,
    dispatch,
    to_layer,
)
from .einops import EinMix, Rearrange, Reduce
from .experiment import Experiment
from .layers import (
    Chain,
    Dropout,
    Embedding,
    F,
    LayerNorm,
    Linear,
    test_mode,
    train_mode,
)
from .observers import (
    ObserverBase,
    CompositeObserver,
    DoAtStep0,
    DoEveryNSteps,
    LossLogger,
    StepTimeLogger,
    default_observer,
)

__all__ = [
    # core
    "LayerBase",
    "LayerLike",
    "Learner",
    "PRNG",
    "Param",
    "PyTree",
    "State",
    "Trainer",
    "dispatch",
    "to_layer",
    # einops
    "EinMix",
    "Rearrange",
    "Reduce",
    # experiment
    "Experiment",
    # layers
    "Chain",
    "Dropout",
    "Embedding",
    "F",
    "Linear",
    "LayerNorm",
    "test_mode",
    "train_mode",
    # observers
    "ObserverBase",
    "CompositeObserver",
    "DoAtStep0",
    "DoEveryNSteps",
    "LossLogger",
    "StepTimeLogger",
    "default_observer",
]



from collections import namedtuple
import numpy as np

from julax.einops import Rearrange, Reduce
from einops import _backends, rearrange, reduce
from einops.tests import FLOAT_REDUCTIONS as REDUCTIONS

import pickle

testcase = namedtuple(
    "testcase", ["pattern", "axes_lengths", "input_shape", "wrong_shapes"]
)


rearrangement_patterns = [
    testcase(
        "b c h w -> b (c h w)",
        dict(c=20),
        (10, 20, 30, 40),
        [(), (10,), (10, 10, 10), (10, 21, 30, 40), [1, 20, 1, 1, 1]],
    ),
    testcase(
        "b c (h1 h2) (w1 w2) -> b (c h2 w2) h1 w1",
        dict(h2=2, w2=2),
        (10, 20, 30, 40),
        [(), (1, 1, 1, 1), (1, 10, 3), ()],
    ),
    testcase(
        "b ... c -> c b ...",
        dict(b=10),
        (10, 20, 30),
        [(), (10,), (5, 10)],
    ),
]


def test_rearrange_imperative():
    backend = _backends.JaxBackend()
    print("Test layer for ", backend.framework_name)

    for pattern, axes_lengths, input_shape, wrong_shapes in rearrangement_patterns:
        x = np.arange(np.prod(input_shape), dtype="float32").reshape(input_shape)
        result_numpy = rearrange(x, pattern, **axes_lengths)
        layer = Rearrange(pattern, **axes_lengths)
        for shape in wrong_shapes:
            try:
                layer(backend.from_numpy(np.zeros(shape, dtype="float32")))
            except BaseException:
                pass
            else:
                raise AssertionError("Failure expected")

        # simple pickling / unpickling
        layer2 = pickle.loads(pickle.dumps(layer))
        result1 = backend.to_numpy(layer(backend.from_numpy(x))[0])
        result2 = backend.to_numpy(layer2(backend.from_numpy(x))[0])
        assert np.allclose(result_numpy, result1)
        assert np.allclose(result1, result2)


reduction_patterns = [
    *rearrangement_patterns,
    testcase("b c h w -> b ()", dict(b=10), (10, 20, 30, 40), [(10,), (10, 20, 30)]),
    testcase(
        "b c (h1 h2) (w1 w2) -> b c h1 w1",
        dict(h1=15, h2=2, w2=2),
        (10, 20, 30, 40),
        [(10, 20, 31, 40)],
    ),
    testcase("b ... c -> b", dict(b=10), (10, 20, 30, 40), [(10,), (11, 10)]),
]


def test_reduce_imperative():
    backend = _backends.JaxBackend()
    print("Test layer for ", backend.framework_name)
    for reduction in REDUCTIONS:
        for pattern, axes_lengths, input_shape, wrong_shapes in reduction_patterns:
            print(backend, reduction, pattern, axes_lengths, input_shape, wrong_shapes)
            x = np.arange(1, 1 + np.prod(input_shape), dtype="float32").reshape(
                input_shape
            )
            if reduction == "prod":
                # make numbers smaller to avoid overflow
                x = x / x.astype("float64").mean() / 10
            else:
                x = x / x.astype("float64").mean()
            result_numpy = reduce(x, pattern, reduction, **axes_lengths)
            layer = Reduce(pattern, reduction, **axes_lengths)
            for shape in wrong_shapes:
                try:
                    layer(backend.from_numpy(np.zeros(shape, dtype="float32")))
                except BaseException:
                    pass
                else:
                    raise AssertionError("Failure expected")

            # simple pickling / unpickling
            layer2 = pickle.loads(pickle.dumps(layer))
            result1 = backend.to_numpy(layer(backend.from_numpy(x))[0])
            result2 = backend.to_numpy(layer2(backend.from_numpy(x))[0])
            assert np.allclose(result_numpy, result1)
            assert np.allclose(result1, result2)



# /// script
# dependencies = [
#   "julax",
#   "pyarrow",
# ]
#
# [tool.uv.sources]
# julax = { path = "../", editable = true }
# ///

import os
import pickle
import optax
from safetensors import safe_open

import grain
import jax
import jax.numpy as jnp
import numpy as np
from grain._src.core.sharding import ShardByJaxProcess, even_split
from grain.experimental import FlatMapIterDataset, FlatMapTransform, ParquetIterDataset
from jax import Array

from julax.base import Dtype
from julax.core import LayerBase, Learner, Param, State, Trainer, PRNG
from julax.einops import Rearrange
from julax.experiment import Experiment
from julax.layers import (
    Branch,
    Chain,
    Embedding,
    Linear,
    Parallel,
    Repeat,
    Residual,
    RMSNorm,
    Select,
)
from julax.observers import default_observer
from julax.utils import identity


# Adapted from:
# https://github.com/AI-Hypercomputer/maxtext/blob/9204d6bbbf8bb19a05ebed72a55cfec687e0e044/src/MaxText/layers/embeddings.py#L486-L622
# TODO: The real and imaginary part are interleaved. benchmark with the HF
# transformer style (first half as real,  second half as imaginary).
def apply_rotary_emb(
    inputs: jax.Array,
    timescale: jax.Array,
    position: None | jax.Array = None,
    fprop_dtype: Dtype | None = jnp.bfloat16,
) -> jax.Array:
    """Applies LLaMA variant of rotary position embedding.

    Args:
        inputs: The input sequence on which to apply the Rotary position
            embedding. It is assumed of shape [B, S, N, H].
        position: Optional position array [B, S]. Only needed when the sequence
            is packed.

    Returns:
        A jax.Array of shape [B, S, N, H] with rotary position embeddings applied.
    """
    # Ensure input is 4D
    if len(inputs.shape) != 4:
        raise ValueError(
            "Input is assumed to be a rank 4 tensor of shape [B, S, N, H]."
        )
    # Determine positions if not provided
    if position is None:
        seq_length = inputs.shape[1]
        position = jnp.arange(seq_length, dtype=jnp.float32)[jnp.newaxis, :]

    # Calculate sinusoidal input
    position = position[:, :, jnp.newaxis, jnp.newaxis]
    sinusoid_inp = position / timescale

    sin = jnp.sin(sinusoid_inp)
    cos = jnp.cos(sinusoid_inp)

    r, i = jnp.split(inputs, 2, axis=-1)
    pos_r = cos * r - sin * i
    pos_i = sin * r + cos * i
    outputs = jnp.concatenate([pos_r, pos_i], axis=-1)

    if fprop_dtype:
        outputs = outputs.astype(fprop_dtype)

    return outputs


class LLaMARotaryEmbedding(LayerBase):
    embedding_dims: int
    min_timescale: int = 1
    max_timescale: int = 10_000
    cast_as_fprop_dtype: bool = True
    fprop_dtype: Dtype = jnp.bfloat16

    scaling_factor: float = 8.0
    low_freq_factor: float = 1.0
    high_freq_factor: float = 4.0
    original_max_position_embeddings: int = 8192

    def _apply_scaling_factor(self, freq):
        """apply scaling factor to rotary position embedding."""
        low_freq_wavelen = self.original_max_position_embeddings / self.low_freq_factor
        high_freq_wavelen = (
            self.original_max_position_embeddings / self.high_freq_factor
        )
        wavelen = 2 * jnp.pi / freq

        def lower_wavelen(freq):
            return freq

        def bigger_or_equal_wavelen(freq):
            def bigger_wavelen(freq):
                return freq / self.scaling_factor

            def equal_wavelen(freq):
                smooth = (
                    self.original_max_position_embeddings / wavelen
                    - self.low_freq_factor
                ) / (self.high_freq_factor - self.low_freq_factor)
                return (1 - smooth) * freq / self.scaling_factor + smooth * freq

            bigger_wavelen_cond = wavelen > low_freq_wavelen
            return jax.lax.cond(
                bigger_wavelen_cond, bigger_wavelen, equal_wavelen, freq
            )

        lower_wavelen_cond = wavelen < high_freq_wavelen
        return jax.lax.cond(
            lower_wavelen_cond, lower_wavelen, bigger_or_equal_wavelen, freq
        )

    def state(self, rng) -> State:
        half_embedding_dim = self.embedding_dims // 2
        fraction = 2 * jnp.arange(0, half_embedding_dim) / self.embedding_dims
        timescale = (
            self.min_timescale * (self.max_timescale / self.min_timescale) ** fraction
        )
        timescale = 1.0 / jax.vmap(self._apply_scaling_factor)(1.0 / timescale)
        return State(timescale=timescale)

    def forward(self, x: Array, p: Param, s: State) -> tuple[Array, State]:
        return apply_rotary_emb(
            x,
            s["timescale"],
            position=None,
            fprop_dtype=self.fprop_dtype if self.cast_as_fprop_dtype else None,
        ), s


class Tokenize(FlatMapTransform):
    def __init__(self, tokenizer_path: str) -> None:
        super().__init__()
        with open(tokenizer_path, "rb") as f:
            self.tokenizer = pickle.load(f)
            self.bos_token_id = self.tokenizer.encode_single_token("<|bos|>")

    def encode(self, text: str) -> list[int]:
        return [self.bos_token_id] + self.tokenizer.encode_ordinary(text)

    def flat_map(self, element):
        return self.encode(element)

    def get_segment_ids(self, tokens: np.ndarray):
        assert tokens.ndim == 2
        bos_mask = tokens == self.bos_token_id
        bos_mask[:, 0] = False
        segment_ids = np.cumsum(bos_mask, axis=1)

        return segment_ids


def create_dataset(
    batch_size: int,
    seq_len: int,
    data_dir: str,
    tokenizer_path: str,
    split: str = "train",
    seed: int = 2025,
) -> grain.IterDataset:
    files = sorted([os.path.join(data_dir, f) for f in os.listdir(data_dir)])
    if split == "train":
        files = files[:-1]
    else:
        # TODO:
        raise ValueError("Unsupported yet")

    tokenize = Tokenize(tokenizer_path)

    # TODO: window shuffle?
    # TODO: prefetch to device?
    ds = grain.experimental.InterleaveIterDataset(
        grain.MapDataset.source(files)
        .shuffle(seed=seed)
        .slice(slice(*even_split(len(files), ShardByJaxProcess(drop_remainder=True))))
        .map(
            lambda file_path: FlatMapIterDataset(
                ParquetIterDataset(file_path).map(lambda x: x["text"]), tokenize
            )
            .batch(batch_size * seq_len + 1)
            .map(np.array)
            .map(
                lambda x: {
                    "inputs": {
                        "token_ids": x[:-1].reshape(batch_size, seq_len),
                        "segment_ids": tokenize.get_segment_ids(
                            x[:-1].reshape(batch_size, seq_len)
                        ),
                    },
                    "target_labels": x[1:].reshape(batch_size, seq_len),
                }
            )
        ),  # pyright: ignore[reportArgumentType]
        cycle_length=4,
    )

    return ds.mp_prefetch(
        grain.MultiprocessingOptions(num_workers=4, per_worker_buffer_size=1)
    )


def attention(inputs):
    q, k, v = inputs["hidden"].values()
    q = apply_rotary_emb(q, inputs["timescale"])
    k = apply_rotary_emb(k, inputs["timescale"])
    o = jax.nn.dot_product_attention(q, k, v, is_causal=True)
    return o


class CachedAttention(LayerBase):
    batch_size: int
    cache_size: int
    num_kv_heads: int
    head_dim: int
    dtype: Dtype = jnp.bfloat16

    def state(self, rng: PRNG) -> State:
        return State(
            k=jnp.zeros(
                (self.batch_size, self.cache_size, self.num_kv_heads, self.head_dim),
                dtype=self.dtype,
            ),
            v=jnp.zeros(
                (self.batch_size, self.cache_size, self.num_kv_heads, self.head_dim),
                dtype=self.dtype,
            ),
            end_index=jnp.zeros(1, dtype=jnp.int32),
        )

    def state_length(self) -> int:
        return (
            2 * self.batch_size * self.cache_size * self.num_kv_heads * self.head_dim
            + 1
        )

    def forward(self, inputs: dict, p: Param, s: Param) -> tuple[Array, State]:
        q, k, v = inputs["hidden"].values()
        seq_len = q.shape[1]

        timescale = inputs["timescale"]
        position = inputs["position"]

        q = apply_rotary_emb(q, timescale, position)
        k = apply_rotary_emb(k, timescale, position)

        slice_indices = (0, s["end_index"][0], 0, 0)
        k = jax.lax.dynamic_update_slice(s["k"], k, slice_indices)
        v = jax.lax.dynamic_update_slice(s["v"], v, slice_indices)
        # o = jax.nn.dot_product_attention(q, k, v, is_causal=True)
        query_positions = jnp.arange(seq_len) + s["end_index"][0]
        key_positions = jnp.arange(self.cache_size)
        attention_mask = key_positions[None, :] <= query_positions[:, None]
        o = jax.nn.dot_product_attention(q, k, v, mask=attention_mask[None, None, :, :])

        S = State(
            k=k,
            v=v,
            end_index=s["end_index"] + seq_len,
        )
        return o, S


class Transformer(LayerBase):
    emb: Embedding
    rope: LLaMARotaryEmbedding
    blocks: Repeat
    out_norm: RMSNorm

    def forward(self, x: dict, p: Param, s: State) -> tuple[Array, State]:
        S = State(rope=s["rope"])

        h = x["token_ids"]
        h, S["emb"] = self.emb(h, p["emb"], s["emb"])
        h, S["blocks"] = self.blocks(
            {
                "hidden": h,
                "timescale": s["rope"]["timescale"],
                "position": x.get("position", None),
            },
            p["blocks"],
            s["blocks"],
        )
        h, S["out_norm"] = self.out_norm(h["hidden"], p["out_norm"], s["out_norm"])

        o = self.emb.attend(h, p["emb"])

        return o, S


def create_transformer(
    batch_size=1,
    seq_len=10,
    dim=2048,
    num_q_heads=32,
    num_kv_heads=8,
    head_dim=64,
    ffn_hidden_dim=8192,
    vocab_size=128256,
    cache_size=None,
) -> Transformer:
    return Transformer(
        emb=Embedding(in_dim=vocab_size, out_dim=dim, param_dtype=jnp.bfloat16),
        rope=LLaMARotaryEmbedding(
            embedding_dims=head_dim,
            min_timescale=1,
            max_timescale=500_000,
        ),
        out_norm=RMSNorm(dim=dim, eps=1e-05, scale_dtype=jnp.bfloat16),
        blocks=Repeat(
            n=16,
            layer=Branch(
                hidden=Chain(
                    attn=Residual(
                        Chain(
                            Parallel(
                                hidden=Chain(
                                    norm=RMSNorm(dim=dim, eps=1e-05),
                                    qkv_proj=Branch(
                                        q=Chain(
                                            Linear(
                                                in_dim=dim,
                                                out_dim=num_q_heads * head_dim,
                                                param_dtype=jnp.bfloat16,
                                            ),
                                            Rearrange(
                                                "B T (N H) -> B T N H",
                                                B=batch_size,
                                                T=seq_len,
                                                N=num_q_heads,
                                                H=head_dim,
                                            ),
                                        ),
                                        k=Chain(
                                            Linear(
                                                in_dim=dim,
                                                out_dim=num_kv_heads * head_dim,
                                                param_dtype=jnp.bfloat16,
                                            ),
                                            Rearrange(
                                                "B S (K H) -> B S K H",
                                                B=batch_size,
                                                S=seq_len,
                                                K=num_kv_heads,
                                                H=head_dim,
                                            ),
                                        ),
                                        v=Chain(
                                            Linear(
                                                in_dim=dim,
                                                out_dim=num_kv_heads * head_dim,
                                                param_dtype=jnp.bfloat16,
                                            ),
                                            Rearrange(
                                                "B S (K H) -> B S K H",
                                                B=batch_size,
                                                S=seq_len,
                                                K=num_kv_heads,
                                                H=head_dim,
                                            ),
                                        ),
                                    ),
                                ),
                                timescale=identity,
                                position=identity,
                                reduce=attention
                                if cache_size is None
                                else CachedAttention(
                                    batch_size=batch_size,
                                    cache_size=cache_size,
                                    num_kv_heads=num_kv_heads,
                                    head_dim=head_dim,
                                    dtype=jnp.bfloat16,
                                ),
                            ),
                            Rearrange(
                                "B T N H -> B T (N H)",
                                B=batch_size,
                                T=seq_len,
                                N=num_q_heads,
                                H=head_dim,
                            ),
                            Linear(
                                in_dim=dim,
                                out_dim=dim,
                                param_dtype=jnp.bfloat16,
                            ),
                        ),
                        skip_through=Select(key="hidden"),
                    ),
                    ffn=Residual(
                        Chain(
                            norm=RMSNorm(dim=dim, eps=1e-05),
                            up=Branch(
                                up_proj=Linear(
                                    in_dim=dim,
                                    out_dim=ffn_hidden_dim,
                                    param_dtype=jnp.bfloat16,
                                ),
                                gate_proj=Chain(
                                    proj=Linear(
                                        in_dim=dim,
                                        out_dim=ffn_hidden_dim,
                                        param_dtype=jnp.bfloat16,
                                    ),
                                    activation=jax.nn.silu,
                                ),
                                reduce=lambda x: x["up_proj"] * x["gate_proj"],
                            ),
                            down=Linear(
                                in_dim=ffn_hidden_dim,
                                out_dim=dim,
                                param_dtype=jnp.bfloat16,
                            ),
                        )
                    ),
                ),
                timescale=Select(key="timescale"),
                position=Select(key="position"),
            ),
        ),
    )


def from_hf(p, s, model_path=None):
    # Allow overriding the default model path via argument or environment variable.
    if model_path is None:
        model_path = os.getenv(
            "LLAMA_MODEL_PATH", "models/Llama-3.2-1B-Instruct/model.safetensors"
        )

    tensors = {}
    with safe_open(model_path, framework="flax", device="cpu") as f:
        for key in f.keys():
            tensors[key] = f.get_tensor(key)

    w_ln1 = []
    w_q = []
    w_k = []
    w_v = []
    w_o = []
    w_ln2 = []
    w_up = []
    w_gate = []
    w_down = []

    for i in range(16):
        w_ln1.append(tensors[f"model.layers.{i}.input_layernorm.weight"])
        w_q.append(tensors[f"model.layers.{i}.self_attn.q_proj.weight"].T)
        w_k.append(tensors[f"model.layers.{i}.self_attn.k_proj.weight"].T)
        w_v.append(tensors[f"model.layers.{i}.self_attn.v_proj.weight"].T)
        w_o.append(tensors[f"model.layers.{i}.self_attn.o_proj.weight"].T)

        w_ln2.append(tensors[f"model.layers.{i}.post_attention_layernorm.weight"])
        w_up.append(tensors[f"model.layers.{i}.mlp.up_proj.weight"].T)
        w_gate.append(tensors[f"model.layers.{i}.mlp.gate_proj.weight"].T)
        w_down.append(tensors[f"model.layers.{i}.mlp.down_proj.weight"].T)

    p["blocks"]["hidden"]["attn"]["processor"]["#0"]["hidden"]["norm"]["scale"] = (
        jnp.stack(w_ln1, axis=0)
    )
    p["blocks"]["hidden"]["attn"]["processor"]["#0"]["hidden"]["qkv_proj"]["q"]["#0"][
        "w"
    ] = jnp.stack(w_q, axis=0)
    p["blocks"]["hidden"]["attn"]["processor"]["#0"]["hidden"]["qkv_proj"]["k"]["#0"][
        "w"
    ] = jnp.stack(w_k, axis=0)
    p["blocks"]["hidden"]["attn"]["processor"]["#0"]["hidden"]["qkv_proj"]["v"]["#0"][
        "w"
    ] = jnp.stack(w_v, axis=0)
    p["blocks"]["hidden"]["attn"]["processor"]["#2"]["w"] = jnp.stack(w_o, axis=0)

    p["blocks"]["hidden"]["ffn"]["processor"]["norm"]["scale"] = jnp.stack(
        w_ln2, axis=0
    )
    p["blocks"]["hidden"]["ffn"]["processor"]["up"]["up_proj"]["w"] = jnp.stack(
        w_up, axis=0
    )
    p["blocks"]["hidden"]["ffn"]["processor"]["up"]["gate_proj"]["proj"]["w"] = (
        jnp.stack(w_gate, axis=0)
    )
    p["blocks"]["hidden"]["ffn"]["processor"]["down"]["w"] = jnp.stack(w_down, axis=0)

    p["emb"]["w"] = tensors["model.embed_tokens.weight"]
    p["out_norm"]["scale"] = tensors["model.norm.weight"]

    return p, s


def verify(max_seq_len=30):
    tokens = [128000, 791, 6367, 311, 28915, 264, 1695, 19692, 374, 220]
    input_ids = jnp.array([tokens])
    m_prefill = create_transformer(seq_len=input_ids.shape[1], cache_size=max_seq_len)
    m_decode = create_transformer(seq_len=1, cache_size=max_seq_len)
    p, s = from_hf(*m_prefill.init())
    o, s_cached = m_prefill(
        {
            "token_ids": input_ids,
            "position": jnp.arange(input_ids.shape[1]).reshape(1, -1),
        },
        p,
        s,
    )

    for _ in range(max_seq_len - 1 - input_ids.shape[1]):
        new_token = int(o[0][-1].argmax())
        position = jnp.array([[len(tokens)]])
        tokens.append(new_token)

        input_ids = jnp.array([[new_token]])

        o, s_cached = m_decode(
            {"token_ids": input_ids, "position": position}, p, s_cached
        )
    print("Generated tokens:", tokens)


def create_experiment():
    return Experiment(
        name="llama_3.2_1b",
        max_steps=1000,
        trainer=Trainer(
            learner=Learner(
                feature_name="inputs",
                label_name="target_labels",
                model=create_transformer(),
                loss_fn=optax.softmax_cross_entropy_with_integer_labels,
            ),
            optimizer=optax.chain(
                optax.clip_by_global_norm(1.0),
                optax.scale_by_adam(
                    b1=0.9,
                    b2=0.95,
                    eps=1e-8,
                ),
                optax.add_decayed_weights(0.1),
                optax.scale_by_schedule(
                    optax.warmup_cosine_decay_schedule(
                        init_value=0.0,
                        peak_value=0.0005,
                        warmup_steps=2_000,
                        decay_steps=30_000,
                        end_value=5.0e-05,
                    )
                ),
                optax.scale(-1.0),
            ),
        ),
        dataset=create_dataset(
            batch_size=4,
            seq_len=4096,
            data_dir="data/wikipedia_parquet",
            tokenizer_path="models/llama_tokenizer.pkl",
        ),
        observer=default_observer(),
    )



from typing import Callable

import jax
import jax.numpy as jnp
from jax import Array
from jax.nn.initializers import Initializer, lecun_normal, ones, variance_scaling, zeros
from jax.sharding import PartitionSpec as P

from julax.base import Dtype, OutShardingType
from julax.utils import identity


from typing import Annotated

from pydantic import Field
from .core import PRNG, LayerBase, LayerLike, Param, PyTree, State, dispatch


class F(LayerBase):
    f: Callable

    def forward(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]:
        return self.f(x), s


@dispatch
def to_layer(x: Callable):
    return F(f=x)


# TODO: generalize to select subtree
class Select(LayerBase):
    key: int | str

    def forward(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]:
        match self.key:
            case int(k):
                return x[k], s
            case str(k) if k.startswith("."):
                return getattr(x, k), s
            case str(k):
                return x[k], s
            case _:
                raise ValueError(f"Unsupported key type: {type(self.key)}")


class Repeat(LayerBase):
    n: int
    layer: LayerLike

    def sublayers(self) -> dict:
        return {f"#{i}": self.layer for i in range(self.n)}

    @dispatch
    def init(self, rng: PRNG) -> tuple[Param, State]:
        def scan_init(carry, rng):
            p, s = self.layer.init(rng)
            return carry, (p, s)

        rngs = jax.random.split(rng, self.n)
        _, (P, S) = jax.lax.scan(scan_init, None, rngs)
        return P, S

    def __getitem__(self, key) -> LayerBase:
        return self.layer

    def forward(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]:
        def scan_forward(x, ps):
            return self.layer(x, *ps)

        o, s = jax.lax.scan(scan_forward, x, (p, s))

        return o, s


class NamedLayers(LayerBase):
    names: Annotated[tuple[str, ...], Field(repr=False)]
    layers: Annotated[tuple[LayerLike, ...], Field(repr=False)]

    def sublayers(self) -> dict:
        return {k: v for k, v in zip(self.names, self.layers)}


class Chain(NamedLayers):
    def __init__(self, *args, **kwargs):
        names = tuple(f"#{i}" for i in range(len(args))) + tuple(kwargs.keys())
        layers = tuple(args) + tuple(kwargs.values())
        super().__init__(names=names, layers=layers)

    def __getitem__(self, key: str | int) -> LayerBase:
        if isinstance(key, int):
            return self.layers[key]
        else:
            return self.sublayers()[key]

    def forward(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]:
        h = x
        S = State()
        for name, layer in zip(self.names, self.layers):
            h, S[name] = layer(h, p[name], s[name])
        return h, S


class Branch(Chain):
    """1 -> N"""

    # place holder to bypass lint check
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]:
        O = {}
        S = State()
        reduce = None
        for name, layer in zip(self.names, self.layers):
            if "reduce" == name:
                reduce = layer
                continue
            O[name], S[name] = layer(x, p[name], s[name])

        if reduce is None:
            return O, S
        else:
            O, S["reduce"] = reduce(O, p["reduce"], s["reduce"])
            return O, S


class Residual(Branch):
    def __init__(self, processor, *, skip_through=identity, reduce: Callable = jnp.add):
        super().__init__(
            processor=processor,
            skip_through=skip_through,
            reduce=lambda x: reduce(x["processor"], x["skip_through"]),
        )


class Parallel(Branch):
    """N -> N"""

    # place holder to bypass lint check
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]:
        reduce = None
        if "reduce" in self.names:
            assert len(x) == len(self.layers) - 1, (
                "Number of inputs must match number of layers."
            )
            assert self.names.index("reduce") == len(self.layers) - 1, (
                "`reduce` layer must be the last layer."
            )
            names = self.names[:-1]
            layers = self.layers[:-1]
            reduce = self.layers[-1]
        else:
            names = self.names
            layers = self.layers
            assert len(x) == len(self.layers), (
                "Number of inputs must match number of layers."
            )

        if isinstance(x, dict):
            inputs = [x[name] for name in names]
        elif isinstance(x, (list, tuple)):
            inputs = x
        else:
            raise ValueError("Input to Parallel must be a dict, list, or tuple.")

        O = {}
        S = State()
        for name, layer, xᵢ in zip(names, layers, inputs, strict=True):
            O[name], S[name] = layer(xᵢ, p[name], s[name])

        if reduce is None:
            return O, S
        else:
            O, S["reduce"] = reduce(O, p["reduce"], s["reduce"])
            return O, S


#####


class Linear(LayerBase):
    in_dim: int
    out_dim: int
    w_init: Initializer = lecun_normal()
    b_init: Initializer | None = None

    param_dtype: Dtype | None = None
    param_sharding: OutShardingType = None
    out_sharding: OutShardingType = None

    def param(self, rng: PRNG) -> Param:
        p = Param()
        rng_w, rng_b = jax.random.split(rng)
        p["w"] = self.w_init(
            rng_w,
            (self.in_dim, self.out_dim),
            dtype=self.param_dtype,
            out_sharding=self.param_sharding,
        )
        if self.b_init:
            p["b"] = self.b_init(
                rng_b,
                (self.out_dim,),
                dtype=self.param_dtype,
                out_sharding=(
                    None if self.param_sharding is None else P(self.param_sharding[-1])
                ),
            )
        return p

    def param_length(self) -> int:
        return self.in_dim * self.out_dim + (self.out_dim if self.b_init else 0)

    def forward(self, x: Array, p: Param, s: State) -> tuple[Array, State]:
        o = jnp.einsum("...d,dh->...h", x, p["w"], out_sharding=self.out_sharding)
        if self.b_init is not None:
            o += p["b"]
        return o, s


class Dropout(LayerBase):
    rate: float

    def state(self, rng: PRNG) -> State:
        return State(rng=rng, is_training=True)

    def state_length(self) -> int:
        return 4  # typically 32 bits?

    def forward(self, x: Array, p: Param, s: State) -> tuple[Array, State]:
        rng, s["rng"] = jax.random.split(s["rng"])
        if s["is_training"] and self.rate > 0:
            mask = jax.random.bernoulli(rng, self.rate, x.shape)
            o = jnp.where(mask, 0, x) / (1 - self.rate)
        else:
            o = x
        return o, s


def _update_mode(s: State, key: str, val):
    return jax.tree.map_with_path(
        lambda path, x: (
            val if jax.tree_util.keystr(path[-1:], simple=True) == key else True
        ),
        s,
    )


def train_mode(s: State):
    return _update_mode(s, "is_training", True)


def test_mode(s: State):
    return _update_mode(s, "is_training", False)


#####


class Embedding(LayerBase):
    in_dim: int
    out_dim: int
    w_init: Initializer = variance_scaling(1.0, "fan_out", "normal")

    param_dtype: Dtype | None = None
    param_sharding: OutShardingType = None
    out_sharding: OutShardingType = None

    def param(self, rng: PRNG) -> Param:
        return Param(
            w=self.w_init(
                rng,
                (self.in_dim, self.out_dim),
                dtype=self.param_dtype,
                out_sharding=self.param_sharding,
            )
        )

    def param_length(self) -> int:
        return self.in_dim * self.out_dim

    def forward(self, x: Array, p: Param, s: State) -> tuple[Array, State]:
        return p["w"].at[x].get(out_sharding=self.out_sharding), s

    def attend(self, x: Array, p: Param) -> Array:
        return jnp.einsum("...ld,nd->...ln", x, p["w"], out_sharding=self.out_sharding)


class RotaryEmbedding(LayerBase):
    """Rotary Position Embedding."""

    # Adapted from https://github.com/AI-Hypercomputer/maxtext/blob/9204d6bbbf8bb19a05ebed72a55cfec687e0e044/src/MaxText/layers/embeddings.py#L271C11-L356C17
    embedding_dims: int
    min_timescale: int = 1
    max_timescale: int = 10000
    cast_as_fprop_dtype: bool = True
    fprop_dtype: Dtype = jnp.bfloat16
    rope_linear_scaling_factor: float = 1.0

    def state(self, rng: PRNG) -> State:
        half_embedding_dim = self.embedding_dims // 2
        fraction = 2 * jnp.arange(0, half_embedding_dim) / self.embedding_dims
        timescale = (
            self.min_timescale * (self.max_timescale / self.min_timescale) ** fraction
        )
        if self.rope_linear_scaling_factor != 1.0:
            timescale = timescale * self.rope_linear_scaling_factor
        return State(timescale=timescale)

    def state_length(self) -> int:
        return self.embedding_dims // 2

    def forward(self, x: Array, p: Param, s: State) -> tuple[Array, State]:
        seq_length = x.shape[1]
        position = jnp.arange(seq_length, dtype=jnp.float32)[
            jnp.newaxis, :, jnp.newaxis, jnp.newaxis
        ]
        sinusoid_inp = position / s["timescale"]
        sin = jnp.sin(sinusoid_inp).astype(x.dtype)
        cos = jnp.cos(sinusoid_inp).astype(x.dtype)
        first_half, second_half = jnp.split(x, 2, axis=-1)
        first_part = first_half * cos - second_half * sin
        second_part = second_half * cos + first_half * sin
        if self.cast_as_fprop_dtype:
            first_part = first_part.astype(self.fprop_dtype)
            second_part = second_part.astype(self.fprop_dtype)
        x_out = jnp.concatenate((first_part, second_part), axis=-1)
        return x_out, s


class LayerNorm(LayerBase):
    dim: int
    epsilon: float = 1e-5
    w_init: Initializer = ones
    b_init: Initializer = zeros
    compute_dtype: Dtype | None = None

    param_dtype: Dtype | None = None
    param_sharding: OutShardingType = None
    out_sharding: OutShardingType = None

    def param(self, rng: PRNG) -> Param:
        w_rng, b_rng = jax.random.split(rng)
        return Param(
            w=self.w_init(
                w_rng,
                (self.dim,),
                dtype=self.param_dtype,
                out_sharding=self.param_sharding,
            ),
            b=self.b_init(
                b_rng,
                (self.dim,),
                dtype=self.param_dtype,
                out_sharding=(
                    None if self.param_sharding is None else P(self.param_sharding[-1])
                ),
            ),
        )

    def param_length(self) -> int:
        return 2 * self.dim

    def forward(self, x: Array, p: Param, s: State) -> tuple[Array, State]:
        x_std = jax.nn.standardize(
            x.astype(self.compute_dtype), epsilon=self.epsilon
        ).astype(self.param_dtype)
        o = x_std * p["w"] + p["b"]
        if self.out_sharding is not None:
            o = jax.lax.with_sharding_constraint(o, self.out_sharding)
        return o, s


class RMSNorm(LayerBase):
    dim: int
    eps: float = 1e-8
    zero_center: bool = False
    scale_init: Initializer | None = ones
    scale_dtype: Dtype | None = None
    scale_sharding: OutShardingType = None

    dtype: Dtype = jnp.float32
    param_sharding: OutShardingType = None
    out_sharding: OutShardingType = None

    def param(self, rng: PRNG) -> Param:
        if self.scale_init is None:
            return Param()
        else:
            return Param(
                scale=self.scale_init(
                    rng,
                    (self.dim,),
                    dtype=self.scale_dtype,
                    out_sharding=(
                        None
                        if self.param_sharding is None
                        else P(self.param_sharding[-1])
                    ),
                )
            )

    def param_length(self) -> int:
        if self.scale_init is None:
            return 0
        else:
            return self.dim

    def forward(self, x: Array, p: Param, s: State) -> tuple[Array, State]:
        x_dtype = x.dtype

        x = x.astype(self.dtype)
        rms = jax.lax.rsqrt(jnp.mean(jnp.square(x), axis=-1, keepdims=True) + self.eps)

        if self.zero_center:
            x = x - x.mean(axis=-1, keepdims=True)

        o = x * rms

        if self.scale_init is not None:
            o = o * p["scale"]

        o = o.astype(x_dtype)

        if self.out_sharding is not None:
            o = jax.lax.with_sharding_constraint(o, self.out_sharding)
        return o, s



# /// script
# dependencies = [
#   "julax",
# ]
#
# [tool.uv.sources]
# julax = { path = "../", editable = true }
# ///

# Reproduce https://sdbuchanan.com/blog/jax-2/

from functools import partial

import grain
import jax
import jax.numpy as jnp
import numpy as np
import optax
from jax.nn.initializers import truncated_normal

from julax.core import Learner, Trainer
from julax.einops import Rearrange
from julax.experiment import Experiment
from julax.layers import (
    Chain,
    Embedding,
    LayerNorm,
    Linear,
    Parallel,
    Repeat,
    Residual,
    RotaryEmbedding,
)
from julax.observers import default_observer
from julax.utils import identity


class FakeSource(grain.sources.RandomAccessDataSource):
    def __init__(self, seq_len: int = 256) -> None:
        self._seq_len = seq_len
        self._data = np.array(
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 8, 7, 6, 5, 4, 3, 2, 1] * 1024
        )

    def __getitem__(self, index: int):
        return {
            "input_ids": self._data[index : index + self._seq_len],
            "target_labels": self._data[index + 1 : index + 1 + self._seq_len],
        }

    def __len__(self) -> int:
        return len(self._data) - self._seq_len


def main(
    seed: int = 5,
    seq_len: int = 256,
    global_batch_size: int = 128,
    num_steps: int = 1000,
    num_vocab: int = 10,
    dim: int = 768,
    num_heads: int = 12,
    head_dim: int = 64,
    num_layers: int = 2,
    param_std: float = 0.02,
):
    return Experiment(
        name="mini_transformer",
        trainer=Trainer(
            learner=Learner(
                feature_name="input_ids",
                label_name="target_labels",
                model=Chain(
                    emb=Embedding(
                        in_dim=num_vocab,
                        out_dim=dim,
                        w_init=truncated_normal(stddev=param_std),
                    ),
                    blocks=Repeat(
                        n=num_layers,
                        layer=Chain(
                            attn=Residual(
                                processor=Chain(
                                    norm_attn=LayerNorm(dim=dim),
                                    attn=Chain(
                                        # qkv projection
                                        Linear(
                                            in_dim=dim,
                                            out_dim=3 * dim,
                                            w_init=truncated_normal(stddev=param_std),
                                            b_init=None,
                                        ),
                                        Rearrange(
                                            "B T (qkv N H) -> B T (qkv N) H",
                                            B=global_batch_size,
                                            T=seq_len,
                                            qkv=3,
                                            N=num_heads,
                                            H=head_dim,
                                        ),
                                        partial(
                                            jnp.split, indices_or_sections=3, axis=2
                                        ),
                                        Parallel(
                                            query=RotaryEmbedding(
                                                embedding_dims=head_dim,
                                                fprop_dtype=jnp.float32,
                                            ),
                                            key=RotaryEmbedding(
                                                embedding_dims=head_dim,
                                                fprop_dtype=jnp.float32,
                                            ),
                                            value=identity,
                                            reduce=partial(
                                                jax.nn.dot_product_attention,
                                                is_causal=True,
                                            ),
                                        ),
                                        Rearrange(
                                            "B T N H -> B T (N H)",
                                            B=global_batch_size,
                                            T=seq_len,
                                            N=num_heads,
                                            H=head_dim,
                                        ),
                                        Linear(
                                            in_dim=dim,
                                            out_dim=dim,
                                            w_init=truncated_normal(stddev=param_std),
                                            b_init=None,
                                        ),
                                    ),
                                )
                            ),
                            mlp=Residual(
                                processor=Chain(
                                    norm_mlp=LayerNorm(dim=dim),
                                    mlp=Chain(
                                        up=Linear(
                                            in_dim=dim,
                                            out_dim=4 * dim,
                                            w_init=truncated_normal(stddev=param_std),
                                            b_init=None,
                                        ),
                                        act=jax.nn.gelu,
                                        down=Linear(
                                            in_dim=4 * dim,
                                            out_dim=dim,
                                            w_init=truncated_normal(stddev=param_std),
                                            b_init=None,
                                        ),
                                    ),
                                )
                            ),
                        ),
                    ),
                    unemb=Linear(
                        in_dim=dim,
                        out_dim=num_vocab,
                        w_init=truncated_normal(stddev=param_std),
                    ),
                ),
                loss_fn=optax.softmax_cross_entropy_with_integer_labels,
            ),
            optimizer=optax.sgd(0.01),
        ),
        dataset=(
            grain.MapDataset.source(FakeSource(seq_len))
            .shuffle(seed=seed)
            .repeat()
            .batch(batch_size=global_batch_size)
            .slice(slice(num_steps))
            .to_iter_dataset()
        ),
        observer=default_observer(),
    )


x = main()
x.run()
x.close()



from typing import Any, Optional
from einops.layers import RearrangeMixin, ReduceMixin
from einops.layers._einmix import _EinmixMixin

import jax
import jax.numpy as jnp
from jax.nn.initializers import Initializer
from pydantic import computed_field

from julax.base import FrozenDict
from .core import LayerBase, Param, PyTree, State, PRNG


class Reduce(LayerBase):
    pattern: str
    reduction: str
    sizes: dict

    def __init__(self, pattern: str, reduction: str, **kwargs):
        super().__init__(pattern=pattern, reduction=reduction, sizes=kwargs)

    def model_post_init(self, context: Any) -> None:
        self._reducer = ReduceMixin(self.pattern, self.reduction, **self.sizes)

    def forward(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]:
        return self._reducer._apply_recipe(x), s


class Rearrange(LayerBase):
    pattern: str
    sizes: FrozenDict

    def __init__(self, pattern: str, **kwargs):
        super().__init__(pattern=pattern, sizes=kwargs)

    def model_post_init(self, context: Any) -> None:
        self._rearranger = RearrangeMixin(self.pattern, **self.sizes)

    def forward(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]:
        return self._rearranger._apply_recipe(x), s


class _Einmix(_EinmixMixin):
    def _create_parameters(self, weight_shape, weight_bound, bias_shape, bias_bound):
        self._w_shape = weight_shape
        self._b_shape = bias_shape

    def _create_rearrange_layers(
        self,
        pre_reshape_pattern: Optional[str],
        pre_reshape_lengths: Optional[dict],
        post_reshape_pattern: Optional[str],
        post_reshape_lengths: Optional[dict],
    ):
        self._pre_rearrange = None
        if pre_reshape_pattern is not None:
            self._pre_rearrange = Rearrange(pre_reshape_pattern, **pre_reshape_lengths)

        self._post_rearrange = None
        if post_reshape_pattern is not None:
            self._post_rearrange = Rearrange(
                post_reshape_pattern, **post_reshape_lengths
            )


class EinMix(LayerBase):
    pattern: str
    w_shape: str
    w_init: Initializer
    b_shape: str | None
    b_init: Initializer | None
    sizes: dict

    @computed_field
    @property
    def einsum_pattern(self) -> str:
        return self._einmix.einsum_pattern

    def __init__(
        self,
        pattern: str,
        *,
        w_shape: str,
        w_init: Initializer,
        b_shape: str | None = None,
        b_init: Initializer | None = None,
        **kwargs,
    ):
        super().__init__(
            pattern=pattern,
            w_shape=w_shape,
            w_init=w_init,
            b_shape=b_shape,
            b_init=b_init,
            sizes=kwargs,
        )

    def model_post_init(self, context: Any) -> None:
        self._einmix = _Einmix(
            pattern=self.pattern,
            weight_shape=self.w_shape,
            bias_shape=self.b_shape,
            **self.sizes,
        )

    def param(self, rng: PRNG) -> Param:
        if self.b_init and self._einmix._b_shape:
            rng_w, rng_b = jax.random.split(rng)
            return Param(
                w=self.w_init(rng_w, self._einmix._w_shape),
                b=self.b_init(rng_b, self._einmix._b_shape),
            )
        else:
            return Param(w=self.w_init(rng, self._einmix._w_shape))

    def forward(self, x: PyTree, p: Param, s: State) -> tuple[PyTree, State]:
        if self._einmix._pre_rearrange is not None:
            x, _ = self._einmix._pre_rearrange(x, p, s)

        o = jnp.einsum(self._einmix.einsum_pattern, x, p["w"])

        if "b" in p:
            o += p["b"]

        if self._einmix._post_rearrange is not None:
            o, _ = self._einmix._post_rearrange(o, p, s)

        return o, s



import math
import jax
from jax.sharding import Mesh
from jax.experimental import mesh_utils


def identity(x):
    return x


def create_mesh(mesh_shape: dict[str, int]) -> Mesh:
    # TODO: support multi-slice
    values = list(mesh_shape.values())
    if -1 in values:
        product = math.prod(v for v in values if v != -1)
        values[values.index(-1)] = jax.device_count() // product

    devices = mesh_utils.create_device_mesh(values, jax.devices())
    return Mesh(devices, list(mesh_shape.keys()))



import jax
from julax.base import dispatch
from julax.core import LayerBase

from rich.console import Console
from rich.tree import Tree
from rich.markup import escape

import humanize
import zlib

from julax.layers import Repeat

console = Console()


def string_color(s: str) -> str:
    return f"color({(zlib.adler32(s.encode()) + 123) % 256})"


def colorof(layer: LayerBase) -> str:
    return string_color(layer.__class__.__name__)


def nameof(layer: LayerBase) -> str:
    s = layer.__class__.__name__
    return f"[{colorof(layer)}]{s}[/]"


def argsof(layer: LayerBase) -> str:
    fields = layer.model_dump(
        exclude_defaults=True, exclude_unset=True, exclude_none=True
    )
    for k, v in type(layer).model_fields.items():
        if k in fields and v.repr is False:
            del fields[k]
        if isinstance(getattr(layer, k), LayerBase):
            del fields[k]

    if fields:
        args = [f"[bright_blue]{k}[/]={escape(repr(v))}" for k, v in fields.items()]
        return "(" + ", ".join(args) + ")"
    else:
        return ""


def param_info(layer: LayerBase) -> str:
    num_params, num_states = layer.numel()
    total = num_params + num_states
    if total == 0:
        return ""
    s = f" [dim]# Total Params: [bright_green]{humanize.intcomma(total)}[/][/]"
    if num_states == 0:
        return s
    else:
        return (
            s
            + f" [trainable=[bright_green]{humanize.intcomma(num_params)}[/], non_trainable=[bright_green]{humanize.intcomma(num_states)}[/]]"
        )


def summary(layer: LayerBase) -> str:
    return f"[bold]{nameof(layer)}[/][dim]{argsof(layer)}[/]"


@dispatch
def to_rich(layer: Repeat) -> Tree:
    root = Tree(summary(layer) + param_info(layer), guide_style=colorof(layer))
    child = to_rich(layer.layer)
    child.label = f"[{colorof(layer.layer)}]0..{layer.n - 1}[/] [bright_yellow]=>[/] {child.label}"
    root.children.append(child)
    return root


@dispatch
def to_rich(layer: LayerBase) -> Tree:
    root = Tree(summary(layer) + param_info(layer), guide_style=colorof(layer))
    for name, sublayer in layer.sublayers().items():
        child = to_rich(sublayer)
        child.label = (
            f"[{colorof(sublayer)}]{name}[/] [bright_yellow]=>[/] {summary(sublayer)}"
        )
        child.guide_style = colorof(sublayer)
        root.children.append(child)
    return root


@dispatch
def to_rich(x: jax.Array) -> str:
    return str(jax.typeof(x))


@dispatch
def to_rich(t: dict) -> Tree:
    root = Tree("")
    for k, v in t.items():
        child = to_rich(v)
        if isinstance(child, Tree):
            if child.label:
                child.label = f"[bright_blue]{k}[/] [bright_yellow]=>[/] {child.label}"
            else:
                child.label = f"[bright_blue]{k}[/]"
            root.children.append(child)
        elif isinstance(child, str):
            root.add(f"[bright_blue]{k}[/] [bright_yellow]=>[/] {child}")
        else:
            raise NotImplementedError()
    return root


def pprint(x):
    console.print(to_rich(x))
